from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.parties.organisation import Organisation


class Company(Organisation):

    def __init__(self, asset_manager_id, party_id, base_currency=None, description='', party_status='Active',
                 addresses=None, emails=None, links=None, references=None,
                 *args, **kwargs):
        if not hasattr(self, 'party_class'):  # A more specific child class may have already set this
            self.party_class = 'Company'
        super(Organisation, self).__init__(asset_manager_id=asset_manager_id, party_id=party_id,
                                           base_currency=base_currency, description=description,
                                           party_status=party_status,
                                           addresses=addresses, emails=emails, links=links, references=references,
                                           *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from amaasutils.random_utils import random_string
import datetime
import random

from amaascore.monitor.item import Item


def generate_item(client_id=None, asset_manager_id=None, item_id=None, item_class=None, item_type=None,
                  item_level=None, item_source=None, item_date=None, message=None):

    item = Item(
        client_id=client_id or random.randint(1, 2**31-1),
        asset_manager_id=asset_manager_id or random.randint(1, 1000),
        item_id=item_id or random_string(10),
        item_class=item_class or random.choice(['Exception', 'Notification']),
        item_type=item_type or random_string(15),
        item_level=item_level or random.choice(['Info', 'Warning', 'Error', 'Critical']),
        item_source=item_source or random.choice(['Transactions', 'Assets', random_string(20)]),
        item_date=item_date or datetime.date.today(),
        message=message or random_string(200)
    )
    return item



from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import uuid

from amaascore.error_messages import ERROR_LOOKUP
from amaascore.books.enums import BOOK_TYPES
from amaascore.core.amaas_model import AMaaSModel


class Book(AMaaSModel):

    @staticmethod
    def non_interface_attributes():
        return ['positions']

    def __init__(self, asset_manager_id, book_id=None, book_type='Trading', book_status='Active', owner_id=None,
                 party_id=None, close_time=datetime.timedelta(hours=18), timezone='UTC', base_currency='USD',
                 business_unit='', description='', positions=None,
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.book_id = book_id or uuid.uuid4().hex
        self.book_type = book_type
        self.book_status = book_status
        self.owner_id = owner_id or party_id  # This could still be None if neither are set, which will raise an error
        self.party_id = party_id
        self.close_time = close_time
        self.timezone = timezone
        self.base_currency = base_currency
        self.business_unit = business_unit
        self.description = description
        self.positions = positions
        super(Book, self).__init__(*args, **kwargs)

    def positions_by_asset(self):
        """
        A dictionary of Position objects keyed by asset_id.
        :return:
        """
        return {position.asset_id: position for position in self.positions}

    @property
    def book_type(self):
        if hasattr(self, '_book_type'):
            return self._book_type

    @book_type.setter
    def book_type(self, book_type):
        """

        :param book_type: The type of book that we are creating - e.g. Trading, Wash
        :return:
        """
        if book_type not in BOOK_TYPES:
            raise ValueError(ERROR_LOOKUP.get('book_type_invalid') % (book_type, self.book_id, self.asset_manager_id))
        else:
            self._book_type = book_type



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date

from amaascore.assets.asset import Asset


class Synthetic(Asset):

    def __init__(self, asset_id, asset_manager_id, asset_issuer_id=None, asset_status='Active',
                 country_id=None, currency=None, description='', fungible=True, issue_date=date.min,
                 maturity_date=date.max, links=None, references=None, *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Synthetic'
        super(Synthetic, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=fungible,
                                        asset_issuer_id=asset_issuer_id, asset_status=asset_status, currency=currency,
                                        issue_date=issue_date, maturity_date=maturity_date, country_id=country_id,
                                        description=description,
                                        links=links, references=references, *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

import csv


def csv_filename_to_objects(filename, json_handler):
    with open(filename, 'r') as f:
        objects = csv_stream_to_objects(f, json_handler=json_handler)
    return objects


def csv_stream_to_objects(stream, json_handler):
    reader = csv.DictReader(stream)
    objects = []
    for row in reader:
        objects.append(json_handler(row))
    return objects


def objects_to_csv(objects, filename, clazz=None):
    with open(filename, 'w') as csvfile:
        objects_to_csv_stream(objects=objects, stream=csvfile, clazz=clazz)


def objects_to_csv_stream(objects, stream, clazz=None):
    if not objects:
        return
    object_dicts = []
    for obj in objects:
        obj_dict = obj.to_json()
        if clazz:
            # FOR NOW - remove all children
            [obj_dict.pop(child, None) for child in clazz.children().keys()]
        object_dicts.append(obj_dict)
    fieldnames = object_dicts[0].keys()
    writer = csv.DictWriter(stream, fieldnames=fieldnames)
    writer.writeheader()
    writer.writerows(object_dicts)



from __future__ import absolute_import, division, print_function, unicode_literals

from dateutil.parser import parse
from decimal import Decimal
import sys

from amaascore.assets.listed_derivative import ListedDerivative

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Future(ListedDerivative):

    def __init__(self, asset_manager_id, asset_id, settlement_type, contract_size, point_value, tick_size,
                 underlying_asset_id=None, quote_unit=None, asset_issuer_id=None, asset_status='Active',
                 issue_date=None, expiry_date=None, description='', country_id=None, venue_id=None,
                 links=None, references=None,
                 *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Future'
        self.settlement_type = settlement_type
        self.contract_size = contract_size
        self.point_value = point_value
        self.tick_size = tick_size
        self.quote_unit = quote_unit
        self.underlying_asset_id = underlying_asset_id
        self.expiry_date = expiry_date
        super(Future, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                     asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                     description=description, country_id=country_id, venue_id=venue_id,
                                     links=links, references=references, issue_date=issue_date, expiry_date=expiry_date,
                                     *args, **kwargs)

    @property
    def settlement_type(self):
        return self._settlement_type

    @settlement_type.setter
    def settlement_type(self, settlement_type):
        if settlement_type:
            if settlement_type in ['Physical', 'Cash']:
                self._settlement_type = settlement_type
            else:
                raise ValueError("Invalid value for settlement_type: %s" % settlement_type)

    @property
    def point_value(self):
        return self._point_value

    @point_value.setter
    def point_value(self, point_value):
        """

        :param point_value: This should be replaced with a calculation
        :return:
        """
        if point_value is not None:
            self._point_value = point_value

    @property
    def tick_size(self):
        return self._tick_size

    @tick_size.setter
    def tick_size(self, tick_size):
        """

        :param tick_size:
        :return:
        """
        if tick_size is not None:
            self._tick_size = Decimal(tick_size)

    @property
    def tick_value(self):
        return self.tick_size * self.contract_size

    @property
    def expiry_date(self):
        if hasattr(self, '_expiry_date'):
            return self._expiry_date

    @expiry_date.setter
    def expiry_date(self, value):
        """
        The date on which the Futures contract expires
        :param expiry_date:
        :return:
        """
        if value:
            self._expiry_date = parse(value).date() if isinstance(value, type_check) else value



from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
from dateutil.parser import parse
import sys
import uuid

from amaascore.core.amaas_model import AMaaSModel
from amaascore.core.reference import Reference

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class CorporateAction(AMaaSModel):

    @staticmethod
    def children():
        return {'references': Reference}

    def __init__(self, asset_manager_id, corporate_action_id, record_date, corporate_action_status='Open',
                 asset_id=None, party_id=None, declared_date=None, settlement_date=None, elective=False, message=None,
                 description='', references=None, *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.corporate_action_id = corporate_action_id or uuid.uuid4().hex
        self.corporate_action_type = self.__class__.__name__
        self.corporate_action_status = corporate_action_status
        self.record_date = record_date
        self.declared_date = declared_date or datetime.date.today()
        self.settlement_date = settlement_date or self.record_date
        self.asset_id = asset_id
        self.party_id = party_id
        self.elective = elective
        self.message = message
        self.description = description
        # Defaults are here not in constructor for mutability reasons.
        self.references = references or {}
        self.references['AMaaS'] = Reference(reference_value=self.corporate_action_id)  # Upserts the AMaaS Reference

        super(CorporateAction, self).__init__(*args, **kwargs)

    @property
    def record_date(self):
        if hasattr(self, '_record_date'):
            return self._record_date

    @record_date.setter
    def record_date(self, value):
        """
        The date on which the corporate action takes effect
        :param value:
        :return:
        """
        if value:
            self._record_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def declared_date(self):
        if hasattr(self, '_record_date'):
            return self._record_date

    @declared_date.setter
    def declared_date(self, value):
        """
        The date on which the corporate action was declared
        :param value:
        :return:
        """
        if value:
            self._declared_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def settlement_date(self):
        if hasattr(self, '_settlement_date'):
            return self._settlement_date

    @settlement_date.setter
    def settlement_date(self, value):
        """
        The date on which the corporate action is settled
        :param value:
        :return:
        """
        if value:
            self._settlement_date = parse(value).date() if isinstance(value, type_check) else value



# POTENTIALLY SUPPORT MULTIPLE ENVIRONMENTS
from __future__ import absolute_import, division, print_function, unicode_literals
LOCAL = False

envs = {'dev', 'staging'}

env = 'dev'

if LOCAL:
    ENDPOINTS = {
        'asset_managers': 'http://localhost:8000',
        'assets': 'http://localhost:8000',
        'books': 'http://localhost:8000',
        'corporate_actions': 'http://localhost:8000',
        'market_data': 'http://localhost:8000',
        'monitor': 'http://localhost:8000',
        'parties': 'http://localhost:8000',
        'transactions': 'http://localhost:8000'
    }
else:
    ENDPOINTS = {
        'asset_managers': 'https://c1hes1s60m.execute-api.ap-southeast-1.amazonaws.com/' + env,
        'assets': 'https://zc6udsq1nb.execute-api.ap-southeast-1.amazonaws.com/' + env,
        'books': 'https://smc367plfg.execute-api.ap-southeast-1.amazonaws.com/' + env,
        'corporate_actions': 'https://basklngdyh.execute-api.ap-southeast-1.amazonaws.com/' + env,
        'market_data': 'https://f0rpi7vksi.execute-api.ap-southeast-1.amazonaws.com/' + env,
        'monitor': 'https://wt50nd7j7l.execute-api.ap-southeast-1.amazonaws.com/' + env,
        'parties': 'https://hpihgzmxoc.execute-api.ap-southeast-1.amazonaws.com/' + env,
        'transactions': 'https://1w0gb581sl.execute-api.ap-southeast-1.amazonaws.com/' + env
    }

DEFAULT_LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s.%(msecs)03d [%(levelname)s] %(name)s: %(message)s',
            'datefmt': '%z %Y-%m-%d %H:%M:%S',
        },
    },
    'handlers': {
        'default': {
            'level': 'INFO',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': True
        },
    }
}



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from amaascore.config import ENDPOINTS
from amaascore.core.interface import Interface
from amaascore.monitor.utils import json_to_item


class MonitorInterface(Interface):

    def __init__(self, logger=None):
        endpoint = ENDPOINTS.get('monitor')
        self.logger = logger or logging.getLogger(__name__)
        super(MonitorInterface, self).__init__(endpoint=endpoint)

    def new_item(self, item):
        url = self.endpoint + '/items'
        response = self.session.post(url, json=item.to_interface())
        if response.ok:
            item = json_to_item(response.json())
            return item
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def resubmit_item(self, asset_manager_id, item_id):
        url = '%s/items/%s/%s' % (self.endpoint, asset_manager_id, item_id)
        response = self.session.patch(url)
        if response.ok:
            item = json_to_item(response.json())
            return item
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_item(self, asset_manager_id, item_id):
        url = '%s/items/%s/%s' % (self.endpoint, asset_manager_id, item_id)
        response = self.session.get(url)
        if response.ok:
            return json_to_item(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def close_item(self, asset_manager_id, item_id):
        url = '%s/items/%s/%s' % (self.endpoint, asset_manager_id, item_id)
        response = self.session.delete(url)
        if response.ok:
            self.logger.info('Successfully Closed Item - Asset Manager: %s - Item ID: %s', asset_manager_id, item_id)
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search_items(self, asset_manager_ids=None, item_ids=None):
        self.logger.info('Search Items - Asset Manager(s): %s', asset_manager_ids)
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if asset_manager_ids:
            search_params['asset_manager_ids'] = asset_manager_ids
        if item_ids:
            search_params['item_ids'] = item_ids
        url = self.endpoint + '/items'
        response = self.session.get(url, params=search_params)
        if response.ok:
            items = [json_to_item(json_item) for json_item in response.json()]
            self.logger.info('Returned %s Items.', len(items))
            return items
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def items_by_asset_manager(self, asset_manager_id):
        self.logger.info('Retrieve Items by Asset Manager: %s', asset_manager_id)
        url = '%s/items/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            items = [json_to_item(json_item) for json_item in response.json()]
            self.logger.info('Returned %s Items.', len(items))
            return items
        else:
            self.logger.error(response.text)
            response.raise_for_status()



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date

from amaascore.assets.asset import Asset


class SyntheticMultiLeg(Asset):
    """
    A synthetic asset which takes multiple assets as 'legs'.  The value of the entire structure is equal to the sum of
    the legs.
    """
    def __init__(self, asset_id, asset_manager_id, legs=None, asset_issuer_id=None, asset_status='Active',
                 country_id=None, currency=None, description='', fungible=True, issue_date=date.min,
                 maturity_date=date.max, links=None, references=None, *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Synthetic'
        self.legs = legs
        super(SyntheticMultiLeg, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                                fungible=fungible, asset_issuer_id=asset_issuer_id,
                                                asset_status=asset_status, currency=currency,
                                                issue_date=issue_date, maturity_date=maturity_date,
                                                country_id=country_id, description=description,
                                                links=links, references=references, *args, **kwargs)

    @property
    def legs(self):
        if hasattr(self, '_legs'):
            return self._legs

    @legs.setter
    def legs(self, legs):
        """
        A list of dictionaries of the legs that make up the multi-legged asset.
        Format is {'asset_id': XYZ, 'quantity': ABC_Decimal}
        :param legs:
        :return:
        """
        if legs is not None:
            if not isinstance(legs, list):
                raise ValueError("Invalid type for asset legs: %s" % type(legs).__name__)
            if not all([isinstance(leg, dict) for leg in legs]):
                raise ValueError("All asset legs must be dictionaries")
            self._legs = legs



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from amaascore.config import ENDPOINTS
from amaascore.core.interface import Interface
from amaascore.corporate_actions.utils import json_to_corporate_action


class CorporateActionsInterface(Interface):

    def __init__(self, logger=None):
        endpoint = ENDPOINTS.get('corporate_actions')
        self.logger = logger or logging.getLogger(__name__)
        super(CorporateActionsInterface, self).__init__(endpoint=endpoint)

    def new(self, corporate_action):
        self.logger.info('New Corporate Action - Asset Manager: %s - Corporate Action ID: %s',
                         corporate_action.asset_manager_id, corporate_action.corporate_action_id)
        url = self.endpoint + '/corporate-actions'
        response = self.session.post(url, json=corporate_action.to_interface())
        if response.ok:
            self.logger.info('Successfully Created Corporate Action - Asset Manager: %s - Corporate Action ID: %s',
                             corporate_action.asset_manager_id, corporate_action.corporate_action_id)
            corporate_action = json_to_corporate_action(response.json())
            return corporate_action
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend(self, corporate_action):
        self.logger.info('Amend Corporate Action - Asset Manager: %s - Corporate Action ID: %s',
                         corporate_action.asset_manager_id, corporate_action.corporate_action_id)
        url = '%s/corporate-actions/%s/%s' % (self.endpoint, corporate_action.asset_manager_id,
                                              corporate_action.corporate_action_id)
        response = self.session.put(url, json=corporate_action.to_interface())
        if response.ok:
            self.logger.info('Successfully Amended Corporate Action - Asset Manager: %s - Corporate Action ID: %s',
                             corporate_action.asset_manager_id, corporate_action.corporate_action_id)
            corporate_action = json_to_corporate_action(response.json())
            return corporate_action
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, corporate_action_id):
        self.logger.info('Retrieve Corporate Action - Asset Manager: %s - Corporate Action ID: %s', asset_manager_id,
                         corporate_action_id)
        url = '%s/corporate-actions/%s/%s' % (self.endpoint, asset_manager_id, corporate_action_id)
        response = self.session.get(url)
        if response.ok:
            self.logger.info('Successfully Retrieved Corporate Action - Asset Manager: %s - Corporate Action ID: %s',
                             asset_manager_id, corporate_action_id)
            return json_to_corporate_action(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def cancel(self, asset_manager_id, corporate_action_id):
        self.logger.info('Cancel Corporate Action - Asset Manager: %s - Corporate Action ID: %s', asset_manager_id,
                         corporate_action_id)
        url = '%s/corporate-actions/%s/%s' % (self.endpoint, asset_manager_id, corporate_action_id)
        json = {'corporate_action_status': 'Cancelled'}
        response = self.session.patch(url, json=json)
        if response.ok:
            self.logger.info('Successfully Cancelled Corporate Action - Asset Manager: %s - Corporate Action ID: %s',
                             asset_manager_id, corporate_action_id)
            return json_to_corporate_action(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_ids=None, corporate_action_ids=None):
        self.logger.info('Search Corporate Actions - Asset Manager(s): %s', asset_manager_ids)
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if asset_manager_ids:
            search_params['asset_manager_ids'] = asset_manager_ids
        if corporate_action_ids:
            search_params['asset_ids'] = corporate_action_ids
        url = self.endpoint + '/corporate-actions'
        response = self.session.get(url, params=search_params)
        if response.ok:
            corp_actions = [json_to_corporate_action(json_corp_action) for json_corp_action in response.json()]
            self.logger.info('Returned %s Corporate Actions.', len(corp_actions))
            return corp_actions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def corporate_actions_by_asset_manager(self, asset_manager_id):
        self.logger.info('Retrieve Corporate Actions by Asset Manager: %s', asset_manager_id)
        url = '%s/corporate-actions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            corp_actions = [json_to_corporate_action(json_corp_action) for json_corp_action in response.json()]
            self.logger.info('Returned %s Corporate Actions.', len(corp_actions))
            return corp_actions
        else:
            self.logger.error(response.text)
            response.raise_for_status()



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from amaascore.config import ENDPOINTS
from amaascore.core.interface import Interface
from amaascore.market_data.utils import json_to_eod_price, json_to_fx_rate


class MarketDataInterface(Interface):

    def __init__(self, logger=None):
        endpoint = ENDPOINTS.get('market_data')
        self.logger = logger or logging.getLogger(__name__)
        super(MarketDataInterface, self).__init__(endpoint=endpoint)

    def persist_eod_prices(self, asset_manager_id, business_date, eod_prices, update_existing_prices=True):
        """

        :param asset_manager_id:
        :param business_date: The business date for which these are rates.  Not really needed, could be derived...
        :param eod_prices:
        :param update_existing_prices:
        :return:
        """
        self.logger.info('Persist EOD Prices - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/eod-prices/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'update_existing_prices': update_existing_prices}
        eod_prices_json = [eod_price.to_interface() for eod_price in eod_prices]
        response = self.session.post(url, params=params, json=eod_prices_json)
        if response.ok:
            eod_prices = [json_to_eod_price(eod_price) for eod_price in response.json()]
            return eod_prices
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_eod_prices(self, asset_manager_id, business_date):
        url = '%s/eod-prices/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        response = self.session.get(url=url)
        if response.ok:
            eod_prices = [json_to_eod_price(eod_price) for eod_price in response.json()]
            self.logger.info('Returned %s EOD Prices.', len(eod_prices))
            return eod_prices
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def persist_fx_rates(self, asset_manager_id, business_date, fx_rates, update_existing_rates=True):
        """

        :param asset_manager_id:
        :param business_date: The business date for which these are rates.  Not really needed, could be derived...
        :param fx_rates:
        :param update_existing_rates:
        :return:
        """
        url = '%s/fx-rates/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'update_existing_rates': update_existing_rates}
        fx_rates_json = [fx_rate.to_interface() for fx_rate in fx_rates]
        response = self.session.post(url, params=params, json=fx_rates_json)
        if response.ok:
            fx_rates = [json_to_fx_rate(fx_rate) for fx_rate in response.json()]
            return fx_rates
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_fx_rates(self, asset_manager_id, business_date):
        url = '%s/fx-rates/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        response = self.session.get(url=url)
        if response.ok:
            fx_rates = [json_to_fx_rate(fx_rate) for fx_rate in response.json()]
            self.logger.info('Returned %s FX Rates.', len(fx_rates))
            return fx_rates
        else:
            self.logger.error(response.text)
            response.raise_for_status()



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.market_data.eod_price import EODPrice
from amaascore.market_data.fx_rate import FXRate


def json_to_eod_price(json_eod_price):
    eod_price = EODPrice(**json_eod_price)
    return eod_price


def json_to_fx_rate(json_fx_rate):
    fx_rate = FXRate(**json_fx_rate)
    return fx_rate




from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.core.amaas_model import AMaaSModel
from amaascore.asset_managers.enums import RELATIONSHIP_TYPES


class Relationship(AMaaSModel):

    def __init__(self, asset_manager_id, relation_id,  relationship_type, client_id, relationship_status='Active',
                 *args, **kwargs):
        """

        :param asset_manager_id: The ID of the Asset Manager who owns this relationship
        :param relation_id: The ID of the Asset Manager to whom this relationship connects
        :param relationship_type: The type of relationship between these Asset Managers
        :param client_id: The client_id that owns this relationship
        :param relationship_status: The status of the relationship
        :param args:
        :param kwargs:
        """
        self.asset_manager_id = asset_manager_id
        self.relation_id = relation_id
        self.relationship_status = relationship_status
        self.relationship_type = relationship_type
        self.client_id = client_id
        super(Relationship, self).__init__(*args, **kwargs)

    @property
    def relationship_type(self):
        if hasattr(self, '_relationship_type'):
            return self._relationship_type

    @relationship_type.setter
    def relationship_type(self, relationship_type):
        if relationship_type:
            if relationship_type not in RELATIONSHIP_TYPES:
                raise ValueError("Invalid Relationship Type: %s" % relationship_type)
            else:
                self._relationship_type= relationship_type



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.transactions.position import Position
from amaascore.transactions.transaction import Transaction


def json_to_position(json_position):
    position = Position(**json_position)
    return position


def json_to_transaction(json_transaction):
    for (collection_name, clazz) in Transaction.children().items():
        children = json_transaction.pop(collection_name, {})
        collection = {}
        for (child_type, child_json) in children.items():
            # Handle the case where there are multiple children for a given type - e.g. links
            if isinstance(child_json, list):
                child = set()
                for child_json_in_list in child_json:
                    child.add(clazz(**child_json_in_list))
            else:
                child = clazz(**child_json)
            collection[child_type] = child
        json_transaction[collection_name] = collection
    transaction = Transaction(**json_transaction)
    return transaction




from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date
from dateutil.parser import parse
import sys
import uuid

from amaascore.core.amaas_model import AMaaSModel
from amaascore.core.reference import Reference

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Asset(AMaaSModel):

    @staticmethod
    def children():
        return {'references': Reference}

    def __init__(self, asset_manager_id, fungible, asset_issuer_id=None, asset_id=None, asset_status='Active',
                 country_id=None, venue_id=None, currency=None, issue_date=date.min, maturity_date=date.max,
                 description='', links=None, references=None,
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.asset_id = asset_id or uuid.uuid4().hex
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Asset'
        self.asset_type = self.__class__.__name__
        self.fungible = fungible
        self.asset_issuer_id = asset_issuer_id
        self.asset_status = asset_status
        self.country_id = country_id
        self.venue_id = venue_id
        self.currency = currency
        self.issue_date = issue_date
        self.maturity_date = maturity_date
        self.description = description
        # Defaults are here not in constructor for mutability reasons.
        self.links = links or {}
        self.references = references or {}
        self.references['AMaaS'] = Reference(reference_value=self.asset_id)  # Upserts the AMaaS Reference

        super(Asset, self).__init__(*args, **kwargs)

    def reference_types(self):
        """
        TODO - are these helper functions useful?
        :return:
        """
        return self.references.keys()

    @property
    def issue_date(self):
        if hasattr(self, '_issue_date'):
            return self._issue_date

    @issue_date.setter
    def issue_date(self, value):
        """
        The date on which the asset is issued
        :param value:
        :return:
        """
        if value:
            self._issue_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def maturity_date(self):
        if hasattr(self, '_maturity_date'):
            return self._maturity_date

    @maturity_date.setter
    def maturity_date(self, value):
        """
        The date on which the asset matures and no longer holds value
        :param value:
        :return:
        """
        if value:
            self._maturity_date = parse(value).date() if isinstance(value, type_check) else value

    def __str__(self):
        return "Asset object - ID: %s" % self.asset_id



from __future__ import absolute_import, division, print_function, unicode_literals


class Portfolio(object):
    """
    TODO - does this derive from anything?  Sort of depends on whether or not we are planning on storing it.
    """

    def __init__(self, books=None):
        self.books = books

    def positions_by_book(self):
        """
        A dictionary of Position objects keyed by book_id.
        :return:
        """
        positions = None
        for book in self.books():
            positions[book] = book.positions()

    def positions_by_asset(self):
        """
        A dictionary of Position objects keyed by asset_id.  If an asset position exists in more than one book, they
         are combined into a single position.
        :return:
        """
        positions = None
        for book in self.books():
            book_positions = book.positions_by_asset()
            # unique positions
            positions.update({position.asset_id: position for position in book_positions
                              if position.asset_id not in positions.keys()})
            # Existing positions
            keys = [position.asset_id for position in book_positions if position.asset_id in positions.key()]
            for key in keys:
                positions.get(key).quantity += book_positions.get(key).quantity
        return positions



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.assets.asset import Asset


class Sukuk(Asset):

    def __init__(self, asset_manager_id, asset_id, maturity_date, asset_issuer_id=None,
                 asset_status='Active',
                 issue_date=None, description='', country_id=None, venue_id=None,
                 links=None, references=None, *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Sukuk'
        super(Sukuk, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=True,
                                    asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                    description=description, country_id=country_id, venue_id=venue_id,
                                    links=links, references=references, issue_date=issue_date,
                                    maturity_date=maturity_date,
                                    *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals


ASSET_MANAGER_TYPES = {'Accredited Investor', 'Bank', 'Broker', 'Corporate Treasury', 'Family Office',
                       'Fund Administrator', 'Fund Manager', 'Hedge Fund', 'Private Equity', 'Retail',
                       'Venture Capital'}
RELATIONSHIP_TYPES = {'Administrator', 'External', 'Front Office', 'Employee'}



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime, date
from dateutil import parser

from amaascore.assets.derivative import Derivative


class ContractForDifference(Derivative):

    def __init__(self, asset_manager_id, asset_id, asset_issuer_id=None, asset_status='Active', description='',
                 country_id=None, venue_id=None, currency=None, issue_date=None, links=None, references=None,
                 *args, **kwargs):
        super(ContractForDifference, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                                    fungible=False, asset_issuer_id=asset_issuer_id,
                                                    asset_status=asset_status, description=description,
                                                    country_id=country_id, venue_id=venue_id, issue_date=issue_date,
                                                    maturity_date=None, currency=currency, links=links,
                                                    references=references,
                                                    *args, **kwargs)
